#!/bin/bash

#
# Description: Script to save power by switching the USB ports of the Raspberry Pi off (and on if needed).
#              based on timetable or internet commands
# Author     : Torsten Tränkner
# License    : GPLv3
# Version    : 2017-04-08
#

# Prerequisites:
#
# 1. add the following line to /etc/sudoers to disable and enable USB
#    pi ALL=(ALL) NOPASSWD: /bin/chmod 666 /sys/devices/platform/soc/3f980000.usb/buspower
#    # 3f980000 depends on the Pi model and operating system
#
#    # tested with hardware: Raspberry Pi 2
#    #         and software: Raspbian Jessie
#
# 2. add the following lines to /etc/sudoers to set the date for correct timer events
#    pi ALL=(ALL) NOPASSWD: /bin/date
#
#    # for test on laptop also:
#    # user ALL=(ALL) NOPASSWD: /sbin/ifconfig
#
# 3. Create a configuration file with your specific needs.

SCRIPT_PATH="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
echo "${SCRIPT_PATH}"

# check if another configuration file for the web server is available
if [ $# -eq 1 ];then
  echo "Using configuration: $1"
  source "$1"
elif [ -e /home/pi/.usb_configuration.txt ];then
  source /home/pi/.usb_configuration.txt
else
  echo "No configuration found. Use a configuration file as first parameter (see documentation)."
  exit 1
fi

if [ "$SUDO_BUS_POWER" == "yes" ];then
  # change the permissions to enable and disable USB as "normal" user
  sudo /bin/chmod 666 "$BUS_POWER"
fi

# constants
USB_DISABLE=0
USB_ENABLE=1

REGULAR_EXPRESSION_NUMBER='^[0-9]{1,5}$'

LIGHT_BLUE="\033[44m\033[1m"
GREEN="\033[42m\033[1m"
RED="\033[41m\033[1m"
ORANGE="\033[43m\033[1m"
MAGENTA="\033[45m\033[1m"
COLOR_END="\033[0m"

DEBUG=1

# global variables
state="on"
usb_disable_seconds=$DEFAULT_SLEEP_SECONDS
last_saved_state="different"
timer_state="on"
internet_state="on"
current_usb_state=$USB_DISABLE
is_active_connection="no"
global_internet_error_counter=0

# variables used to detect an idle connection
start_time=0
end_time=0
active_connection=0

# last uptime seconds since the NTP time was set
last_seconds_since_network_time_was_set=0

# these arrays are calculated to switch the USB ports on and off by time / seconds
declare -a start_seconds
declare -a stop_seconds
declare -a state_array

# create an associative array
declare -A day_of_week_name
for ((i=0; i<${#day_of_week_names[@]};i++));do
  eval "day_of_week_name[\"${day_of_week_names[$i]}\"]=$i"
done

# create a directory to save the states in case of a reboot
if [ ! -d "$STATE_DIRECTORY" ];then
  mkdir "$STATE_DIRECTORY"
fi

touch "$STATE_FILE"

# returns/prints the seconds since start of the pi
function get_uptime_seconds() {
  cat /proc/uptime | sed 's|\..*||'
}

# log to syslog and standard out
function log() {
  logger "$1"

  if [ "${1:0:9}" == "Could not" ];then
    echo -e "${RED}$1${COLOR_END}"
  else
    echo "$1"
  fi
}

# disable and enable the USB ports and wait for the reconnection to the internet after USB ports were enabled
function usb() {
  if [ $current_usb_state -eq $1 ];then
    echo "$(date) USB state not changed $1."
    return
  fi
  current_usb_state=$1

  if [ $DEBUG -gt 0 ];then
    if [ $current_usb_state -eq $USB_DISABLE ];then
      if [ "$(ls -l /sys/class/leds/led1/brightness 2>/dev/null| grep rw-rw-rw)" != "" ];then
        echo 0 > /sys/class/leds/led1/brightness
      fi
      echo -e "$(date) ${RED}USB disabled.${COLOR_END}"
    else
      if [ "$(ls -l /sys/class/leds/led1/brightness 2>/dev/null | grep rw-rw-rw)" != "" ];then
        echo 1 > /sys/class/leds/led1/brightness
      fi
      echo -e "$(date) ${GREEN}USB enabled.${COLOR_END}"
    fi
  fi

  echo "$current_usb_state" > "$BUS_POWER"

  # for test on a linux computer with eth0
  if [ "$SUDO_BUS_POWER" == "no" ];then
    if [ $current_usb_state -eq $USB_ENABLE ];then
      sudo /sbin/ifconfig $MONITOR_INTERFACE up
    else
      sudo /sbin/ifconfig $MONITOR_INTERFACE down
    fi
  fi

  # wait for interface MONITOR_INTERFACE to appear again and get an IP address (e.g. via DHCP)
  # can take up to 8 seconds, depends on your network configuration
  if [ $current_usb_state -eq $USB_ENABLE ];then

    result=1
    interface_up_counter=0
    while [ $result -ne 0 ];do
      ping -c 1 -W 1 "$REFERENCE_IP_ADDRESS_TO_PING" > /dev/null 2>&1
      result=$?
      let interface_up_counter++
      if [ $interface_up_counter -ge $MAXIMUM_RECONNECTION_TIMEOUT ];then
        result=0
      fi
      sleep 0.2
    done
    if [ $DEBUG -gt 0 ];then
      echo -e "$(date) Finished USB enabling."
    fi

    # reset connection parameters
    seconds_since_inactive_connection=0
    is_active_connection="no"
    start_bytes=$(received_bytes)
    inactive_connection_seconds=-1
    echo "$(date) After USB_ENABLE ${start_bytes}"

    # just for testing with local interface which takes some time
    if [ "$SUDO_BUS_POWER" == "no" ];then
      sleep 1
    fi

    let last_seconds_since_active_connection_check=$(get_uptime_seconds)+DEFAULT_ACTIVE_CONNECTION_CHECK_SECONDS

  else
    let last_seconds_since_internet_check_during_off=$(get_uptime_seconds)+usb_disable_seconds
  fi
}

# do not use this function as it might be queried too often
# wget is used instead
function download_state_github() {
  cd "$GITHUB_DIRECTORY"

  # wait some seconds for git fetch
  timeout 10 git fetch --all > /dev/shm/gitpull_$(date +%s).txt 2>&1
  timeout 2 git reset --hard origin/master >> /dev/shm/gitpull_$(date +%s).txt 2>&1
}

function upload_state_github() {
  cd "$GITHUB_DIRECTORY"

  # save state for github
  echo "$1" > state
  rm -rf .git
  git init
  git add .
  git commit -m "Initial commit"
  git remote add origin "$GITHUB_REPOSITORY"

  # upload to github
  git push -u --force origin master
}

# download the state of the Raspberry Pi from the internet / get_state
# and store it in internet_state
function download_internet_state() {

  # download the state
  numberOfRetries=0
  while [ $numberOfRetries -lt $MAXIMUM_INTERNET_STATE_RETRIES ];do

    ##### WEBSERVER #####
    if [ $USE_WEB_SERVER -eq 1 ];then

      # prevent ~/.netrc from being read by wget:
      ORIGINAL_HOME="$HOME"; HOME=
      state=$(wget $WGET_PARAMETERS -T $MAXIMUM_CONNECTION_TIMEOUT $AUTHENTICATION "$STATE_URL" -O -)
      result=$?
      HOME="$ORIGINAL_HOME"

    else
      # download_state_github
      # result=$?
      # state=$(head -n 1 "$GITHUB_DIRECTORY"/state)

      # raw.githubusercontent.com is too slow sometimes

      # prevent ~/.netrc from being read by wget:
      ORIGINAL_HOME="$HOME"; HOME=
      state=$(wget $WGET_PARAMETERS -T $MAXIMUM_CONNECTION_TIMEOUT "$GITHUB_STATE_URL" -O - 2>/dev/null | grep -m 1 LC1 | sed 's|.*">\(.*\)</td>.*|\1|g')
      result=$?
      HOME="$ORIGINAL_HOME"
      if [ "$state" == "" ];then
        result=1
      else
        echo "$state" > "$GITHUB_DIRECTORY/state"
      fi

    fi

    if [ $result -ne 0 ];then
      let numberOfRetries++
      let global_internet_error_counter++
      echo "$(date) Retrying $numberOfRetries. Error counter: $global_internet_error_counter"
      sleep 1
    else
      global_internet_error_counter=0
      numberOfRetries=100
    fi
  done

  if [ $result -ne 0 ];then
    log "Could not download the state from $STATE_URL. Return code: $result."

    # on error set default state to on
    state="on"
  else

    # update the time from the network (something like NTP over HTTP)
    current_time=$(get_uptime_seconds)
    let time_difference=current_time-last_seconds_since_network_time_was_set

    # check if time from internet needs to be set
    if [ $time_difference -ge $MAXIMUM_TIME_SET_DIFFERENCE ];then
      new_time=$(curl -s -I "$TIME_URL" | grep ^Date | sed 's|Date: ||')
      if [ ! -z "$new_time" ];then
        if [ $DEBUG -gt 0 ];then
          echo -e "$(date) ${LIGHT_BLUE}Now setting time.${COLOR_END}"
        else
          log "$(date) Now setting time."
        fi
        sudo date -s "$new_time" > /dev/null
      else
        if [ $DEBUG -gt 0 ];then
          echo -e "$(date) ${LIGHT_BLUE}Setting time failed.${COLOR_END}"
        else
          log "$(date) Setting time failed. No internet."
        fi
      fi

      last_seconds_since_network_time_was_set=$current_time
    fi
  fi

  # translate the state "sus" or "suspend" to the internal state "on suspend"
  if [ "$state" == "sus" ] || [ "$state" == "suspend" ]; then
    state="suspend"
    if [ "$internet_state" == "off" ] || [[ $internet_state =~ $REGULAR_EXPRESSION_NUMBER ]]; then
      is_active_connection="yes"
      start_bytes=$(received_bytes)
    fi
  fi

  # validate state on, off, suspend or sleep seconds up to 99999
  if [ "$state" == "on" ] || [ "$state" == "off" ] || [ "$state" == "suspend" ] || [[ $state =~ $REGULAR_EXPRESSION_NUMBER ]]; then
    # echo "Matched: $state"
    :
  else
    log "Could not validate state: $state"
    state="on"
  fi

  internet_state="$state"

  if [ $global_internet_error_counter -gt $MAXIMUM_INTERNET_ERROR_COUNTER ];then
    global_internet_error_counter=0

    # try to disable and enable USB in error case
    usb $USB_DISABLE
    usb $USB_ENABLE
  fi
}

# upload the USB state to the internet as feedback
# (off or the specific seconds that the Raspberry Pi sleeps between checks)
function upload_state() {
  APPEND="off"
  if [ $1 -ne $DEFAULT_SLEEP_SECONDS ];then
    APPEND="$1"
  fi

  # set internet state
  internet_state=$APPEND

  ##### WEBSERVER #####
  if [ $USE_WEB_SERVER -eq 1 ];then
    ORIGINAL_HOME="$HOME"; HOME=
    wget $WGET_PARAMETERS -T $MAXIMUM_CONNECTION_TIMEOUT $AUTHENTICATION "${OFF_URL}${APPEND}" -O - > /dev/null
    result=$?
    HOME="$ORIGINAL_HOME"
  else
    upload_state_github "${APPEND}"
    result=$?
  fi

  if [ $result -ne 0 ];then
    log "Could not set the state. Result: $result."
  fi
}

# save the IP address of the Raspberry Pi (something like DynDNS)
function set_ip_address() {
  echo -e "$(date) ${MAGENTA}Setting IP address.${COLOR_END}"

  ##### WEBSERVER #####
  if [ $USE_WEB_SERVER -eq 1 ];then
    ORIGINAL_HOME="$HOME"; HOME=
    wget $WGET_PARAMETERS -T $MAXIMUM_CONNECTION_TIMEOUT $AUTHENTICATION "${SET_IP_URL}" -O - > /dev/null
    result=$?
    HOME="$ORIGINAL_HOME"
  else
    "${SCRIPT_PATH}"/free_server/set_dns.sh
    result=$?
  fi

  if [ $result -ne 0 ];then
    log "Could not set the ip address. Result: $result."
  fi
}

# prints the bytes received at the network interface MONITOR_INTERFACE to detect an active network connection
function received_bytes() {

  /sbin/ifconfig $MONITOR_INTERFACE 2>&1 | grep -q "inet "
  result=$?

  # check if the interface that shall be monitored exists
  if [ $result -eq 0 ];then
    /sbin/ifconfig $MONITOR_INTERFACE 2>&1 | grep "RX..ytes" | sed -e 's|.*RX..ytes:||' -e 's| (.*||g'
  else
    echo -n "0"
  fi
}

# get the seconds from monday 0:00 to now
function get_seconds_since_week_start() {
  # check whether today is monday
  if [ $(date +%u) -eq 1 ];then
    week_start_seconds=$(date -d "monday 0:00" +%s)
  else
    week_start_seconds=$(date -d "last monday 0:00" +%s)
  fi
  current_seconds=$(date +%s)

  let seconds_since_week_start=current_seconds-week_start_seconds
  echo "$seconds_since_week_start"
}

# prints the seconds from a time format like "17:35"
function get_seconds_from_time() {
  time="$1"

  # remove a leading zero
  hours=$(  echo "$time" | sed -e 's|:.*||' -e 's|^0||g')
  minutes=$(echo "$time" | sed -e 's|.*:||' -e 's|^0||g')
  let day_seconds=hours*3600+minutes*60
  echo $day_seconds
}

# calculate the seconds from the strings defined in "predefined_time_strings"
# this makes it easier to compare the seconds
function calculate_timer_events_seconds() {

  rm -f /dev/shm/times.txt

  new_index=0

  # convert the strings into an array of seconds
  echo "$predefined_time_strings" | while read line; do

    if [ -z "$line" ] || [ "${line:0:1}" == "#" ];then
      continue
    fi

    echo >> /dev/shm/times.txt

    newString=$(echo "$line" | sed -e 's|" *, *"|"|g' -e 's|^"||' -e 's|"$||')
    # echo "$newString"
    IFS='"' read -r -a temporaryArray <<< "$newString"

    day_of_week_start=${day_of_week_name[${temporaryArray[0]}]}
    day_time_in_seconds=$(get_seconds_from_time ${temporaryArray[1]})
    let start_seconds[new_index]=day_of_week_start*86400+day_time_in_seconds

    day_of_week_end=${day_of_week_name[${temporaryArray[2]}]}
    day_time_in_seconds=$(get_seconds_from_time ${temporaryArray[3]})
    let stop_seconds[new_index]=day_of_week_end*86400+day_time_in_seconds
    state_array[new_index]=${temporaryArray[4]}

    # additional check for times between sunday and monday
    if [ ${stop_seconds[new_index]} -lt ${start_seconds[new_index]} ];then
      old_index=$new_index
      let new_index++

      # add a new entry sunday end and monday start
      state_array[new_index]=${state_array[old_index]}
      stop_seconds[new_index]=${stop_seconds[old_index]}
      stop_seconds[old_index]=604800
      start_seconds[new_index]=0

      echo "start_seconds[$old_index]=${start_seconds[old_index]}" >> /dev/shm/times.txt
      echo "stop_seconds[$old_index]=${stop_seconds[old_index]}" >> /dev/shm/times.txt
      echo "state_array[$old_index]=${state_array[old_index]}" >> /dev/shm/times.txt
      echo >> /dev/shm/times.txt
    fi

    echo "start_seconds[$new_index]=${start_seconds[new_index]}" >> /dev/shm/times.txt
    echo "stop_seconds[$new_index]=${stop_seconds[new_index]}" >> /dev/shm/times.txt
    echo "state_array[$new_index]=${state_array[new_index]}" >> /dev/shm/times.txt

    let new_index++

  done

  # the file needs to be sourced to get the variables into the scope of this script
  source /dev/shm/times.txt
}

# save the current USB state to disk (can be read by external programs)
function save_state() {
  current_state="$1"
  if [ "$last_saved_state" == "$current_state" ];then
    return
  fi
  last_saved_state="$current_state"

  read_state=$(cat "$STATE_FILE")
  if [ "$read_state" != "$current_state" ];then
    echo "$1" > "$STATE_FILE"
    echo "$1" > "$STATE_FILE_MEMORY"
  fi
}

# sets the USB state that is relevant for the current time (timer_state)
function get_timer_state_by_current_time() {
  seconds_since_week_start=$(get_seconds_since_week_start)

  matched="off"

  for ((index = 0; index < ${#state_array[@]}; index++)); do

    # print debug output
    if [ $DEBUG -gt 1 ];then
      echo "Start seconds $index: ${start_seconds[index]} to ${stop_seconds[index]} set to ${state_array[index]}"
    fi

    # check if the time
    if [ $seconds_since_week_start -ge ${start_seconds[index]} ] && [ $seconds_since_week_start -lt ${stop_seconds[index]} ];then
      if [ $DEBUG -gt 1 ];then
        echo "Time matched !!! Start seconds $index: ${start_seconds[index]} to ${stop_seconds[index]} set to ${state_array[index]}"
      fi
      matched="${state_array[index]}"
      break
    fi
  done

  timer_state="$matched"
}

calculate_timer_events_seconds

function check_timer() {
  # check for timer events
  get_timer_state_by_current_time
  counter_timer_check=0

  # set internet check interval by timer event
  if [[ $timer_state =~ $REGULAR_EXPRESSION_NUMBER ]];then
    if [[ $internet_state =~ $REGULAR_EXPRESSION_NUMBER ]];then
      :
    else
      usb_disable_seconds=$timer_state
    fi
  fi

  if [ $DEBUG -gt 0 ];then echo -e "$(date) Timer state: ${GREEN}$timer_state${COLOR_END}. usb_disable_seconds: $usb_disable_seconds"; fi
}

function check_internet() {
  download_internet_state
  if [ $DEBUG -gt 0 ];then
    echo -e "$(date) Internet state: ${ORANGE}$internet_state.${COLOR_END}"
  fi

  # set the number of seconds to sleep
  if [ "$internet_state" == "off" ];then
    if [[ $timer_state =~ $REGULAR_EXPRESSION_NUMBER ]]; then
      echo "$(date) Timer settings have higher priority than default value."
    else
      usb_disable_seconds=$DEFAULT_SLEEP_SECONDS
    fi
  elif [[ $internet_state =~ $REGULAR_EXPRESSION_NUMBER ]];then
    if [[ $timer_state =~ $REGULAR_EXPRESSION_NUMBER ]]; then
      :
    else
      usb_disable_seconds=$internet_state
    fi
  fi

}

# Set the combined state of the USB interface
# composed of timer, internet and active connection
function set_combined_usb_state() {

  # switch USB off if:
  # 1. timer event says off or seconds (not "on")
  # 2. internet says suspend, off or seconds (not "on")
  # 3. no more active connection
  if [ "$timer_state" != "on" ] && [ "$internet_state" != "on" ] && [ "$is_active_connection" == "no" ];then
    combined_usb_state="off"
  else
    combined_usb_state="on"
  fi
}


last_seconds_since_internet_check_during_off=0
last_seconds_since_internet_check_during_on=0
last_seconds_since_ip_upload=0
last_seconds_since_timer_check=0
last_seconds_since_active_connection_check=0
let last_seconds_since_active_connection_check=$(get_uptime_seconds)+DEFAULT_ACTIVE_CONNECTION_CHECK_SECONDS
inactive_connection_seconds=-1

# check for an active connection
function check_active_connection() {
  stop_bytes=$(received_bytes)

  let byte_difference=stop_bytes-start_bytes
  start_bytes=$stop_bytes

  echo -n "$(date) Connection check. Byte difference in the last $DEFAULT_ACTIVE_CONNECTION_CHECK_SECONDS seconds: $byte_difference. "

  if [ $byte_difference -lt $MINIMUM_IDLE_BYTES ];then

    if [ $inactive_connection_seconds -eq -1 ]; then
      inactive_connection_seconds=$(get_uptime_seconds)
      seconds_since_inactive_connection=0
    else
      let seconds_since_inactive_connection=$(get_uptime_seconds)-inactive_connection_seconds
      if [ $seconds_since_inactive_connection -ge $DEFAULT_SUSPEND_SECONDS ];then
        is_active_connection="no"
        seconds_since_inactive_connection=$DEFAULT_SUSPEND_SECONDS
      fi
    fi

    echo -e "${GREEN}No connections${COLOR_END}. Since ${seconds_since_inactive_connection} seconds. Is connection active ? ${is_active_connection}"
    active_connection=0
  else
    inactive_connection_seconds=-1
    seconds_since_inactive_connection=0
    is_active_connection="yes"
    echo -e "${MAGENTA}Active connections${COLOR_END}."
    active_connection=1
  fi
}

# start condition:
# current_usb_state is enabled

# enable USB in case it has been disabled before the start of the script
usb $USB_ENABLE
save_state $USB_ENABLE

# define the number of bytes received at start to detect active connection
start_bytes=$(received_bytes)

# main loop
while true; do

  # check timer events
  if [ $(get_uptime_seconds) -ge $last_seconds_since_timer_check ];then
    check_timer
    if [ "$timer_state" == "on" ] && [ $current_usb_state -eq $USB_DISABLE ];then
      usb $USB_ENABLE
      save_state $USB_ENABLE
      set_ip_address
      let last_seconds_since_internet_check_during_on=$(get_uptime_seconds)+DEFAULT_INTERNET_CHECK_SECONDS_DURING_ON
    fi
    let last_seconds_since_timer_check=$(get_uptime_seconds)+SLEEP_SECONDS_TIMER_CHECK
  fi

  # USB is disabled
  if [ $current_usb_state -eq $USB_DISABLE ];then

    # check only the internet state from time to time
    if [ $(get_uptime_seconds) -ge $last_seconds_since_internet_check_during_off ];then

      # enable USB after the defined timeout to check the internet connection
      # (timeout derived from timer event or internet state)
      usb $USB_ENABLE

      check_internet
      if [ "$internet_state" == "on" ] || [ "$internet_state" == "suspend" ];then
        let last_seconds_since_internet_check_during_on=$(get_uptime_seconds)+DEFAULT_INTERNET_CHECK_SECONDS_DURING_ON
        save_state $USB_ENABLE
        set_ip_address
      else
        ##### USB IS DISABLED HERE
        usb $USB_DISABLE
      fi
    fi

  # USB is enabled
  else

    # check internet state when USB is enabled
    if [ $(get_uptime_seconds) -ge $last_seconds_since_internet_check_during_on ];then
      if [ $(get_uptime_seconds) -ge $last_seconds_since_ip_upload ];then
        set_ip_address
        let last_seconds_since_ip_upload=$(get_uptime_seconds)+DEFAULT_IP_UPLOAD_SECONDS
      fi
      if [ "$timer_state" != "on" ];then
        check_internet
      fi
      let last_seconds_since_internet_check_during_on=$(get_uptime_seconds)+DEFAULT_INTERNET_CHECK_SECONDS_DURING_ON
    fi

    # check for active connection
    if [ $(get_uptime_seconds) -ge $last_seconds_since_active_connection_check ];then
      check_active_connection

      set_combined_usb_state

      ##### USB IS DISABLED INTENTIONALLY HERE
      if [ "$combined_usb_state" == "off" ];then
        if [ "$internet_state" == "suspend" ];then
          upload_state $usb_disable_seconds
        fi

        usb $USB_DISABLE
        save_state $USB_DISABLE
      else
        let last_seconds_since_active_connection_check=$(get_uptime_seconds)+DEFAULT_ACTIVE_CONNECTION_CHECK_SECONDS
      fi

    fi

  fi

  sleep $MINIMAL_SLEEP_TIMEOUT

done
